from __future__ import absolute_import
# import orm here so that event registration work
import sqlalchemy.orm

from abc import abstractmethod
from mortar_import.diff import Diff
from sqlalchemy import inspect


class SQLAlchemyDiff(Diff):

    def __init__(self, session, model, existing, imported):
        super(SQLAlchemyDiff, self).__init__(existing, imported)
        self.model = model
        self.session = session

    def extract_existing(self, obj):
        i = inspect(obj)
        key = i.identity
        extracted = {name: getattr(obj, name) for name in i.attrs.keys()}
        return key, extracted

    @abstractmethod
    def extract_imported(self, obj):
        """
        Must return ``key, dict_`` where ``key`` is the key of the imported
        object and ``dict_`` is a mapping of all keys to values for the
        imported object.
        """

    def add(self, key, imported, extracted_imported):
        self.session.add(self.model(**extracted_imported))

    def update(self,
               key,
               existing, existing_extracted,
               imported, imported_extracted):
        for key, value in imported_extracted.items():
            setattr(existing, key, value)

    def delete(self, key, existing, existing_extracted):
        self.session.delete(existing)



class SingleKeyDictExtractor(object):

    def __init__(self, key):
        self.key = key

    def __call__(self, obj):
        return obj[self.key], obj


class MultiKeyDictExtractor(object):

    def __init__(self, *keys):
        self.keys = keys

    def __call__(self, obj):
        return tuple(obj[key] for key in self.keys), obj


def DictExtractor(*keys):
    if len(keys) == 1:
        return SingleKeyDictExtractor(keys[0])
    else:
        return MultiKeyDictExtractor(*keys)


class NamedTupleExtractor(object):

    def __init__(self, *keys):
        self.keys = keys

    def __call__(self, raw):
        key = tuple(getattr(raw, name) for name in self.keys)
        obj = dict(zip(raw._fields, raw))
        return key, obj


from abc import ABCMeta, abstractmethod
from collections import namedtuple
from six import with_metaclass

Addition = namedtuple('Addition', 'key imported extracted_imported')
Update = namedtuple('Update', 'key '
                              'existing existing_extracted '
                              'imported imported_extracted')
Deletion = namedtuple('Deletion', 'key existing existing_imported')

class Diff(with_metaclass(ABCMeta, object)):

    to_add = to_update = to_delete = None

    def __init__(self, existing, imported):
        self.existing = existing
        self.imported = imported

    @abstractmethod
    def extract_existing(self, obj):
        """
        Must return ``key, obj`` where ``key`` is the key of the existing
        object and ``obj`` is an object that can be compared with
        imported objects.
        """

    @abstractmethod
    def extract_imported(self, obj):
        """
        Must return ``key, obj`` where ``key`` is the key of the imported
        object and ``obj`` is an object that can be compared with
        existing objects.
        """

    @abstractmethod
    def add(self, key, imported, extracted_imported):
        """
        Handle the addition of an imported object.
        """

    @abstractmethod
    def update(self,
               key,
               existing, existing_extracted,
               imported, imported_extracted):
        """
        Handle an update of an existing object.
        """

    @abstractmethod
    def delete(self, key, existing, existing_extracted):
        """
        Handle the deletion of an existing object.
        """

    def compute(self):
        for name in 'existing', 'imported':
            mapping = {}
            sequence = getattr(self, name)
            extract = getattr(self, 'extract_'+name)
            for raw in sequence:
                key, extracted = extract(raw)
                if key in mapping:
                    existing_raw, existing_extracted = mapping[key]
                    raise KeyError(
                        "{key!r} occurs more than once in {name}, "
                        "first was {existing_extracted!r} from {existing!r}, "
                        "next was {new_extracted!r} from {new!r}".format(
                            key=key, name=name,
                            existing_extracted=existing_extracted,
                            existing=existing_raw,
                            new_extracted=extracted,
                            new=raw,
                        ))
                mapping[key] = raw, extracted
            setattr(self, name+'_mapping', mapping)
            setattr(self, name+'_keys', set(mapping))

        self.to_add = []
        self.to_update = []
        self.to_delete = []

        for key in sorted(self.imported_keys - self.existing_keys):
            self.to_add.append(Addition(key, *self.imported_mapping[key]))

        for key in sorted(self.imported_keys & self.existing_keys):
            existing, existing_extracted = self.existing_mapping[key]
            imported, imported_extracted = self.imported_mapping[key]
            if existing_extracted != imported_extracted:
                self.to_update.append(Update(key,
                                             existing, existing_extracted,
                                             imported, imported_extracted))

        for key in sorted(self.existing_keys - self.imported_keys):
            self.to_delete.append(Deletion(key, *self.existing_mapping[key]))

    def apply(self):
        if self.to_add is None:
            self.compute()
        for op in 'add', 'update', 'delete':
            meth = getattr(self, op)
            for action in  getattr(self, 'to_'+op):
                meth(*action)




from unittest import TestCase

from mortar_import.extractors import MultiKeyDictExtractor
from mortar_import.sqlalchemy import SQLAlchemyDiff
from mortar_rdb import get_session
from mortar_rdb.testing import register_session
from sqlalchemy import Column, Integer
from sqlalchemy import String
from sqlalchemy.ext.declarative import declarative_base
from testfixtures import ShouldRaise, compare

Base = declarative_base()


class Simple(Base):
    __tablename__ = 'simple'
    key = Column(String, primary_key=True)
    value = Column(Integer)


class MultiPK(Base):
    __tablename__ = 'multi_pk'
    name = Column(String, primary_key=True)
    index = Column(Integer, primary_key=True)
    value = Column(Integer)


class TestSQLAlchemy(TestCase):

    def setUp(self):
        register_session(transactional=False)
        self.session = get_session()
        self.addCleanup(self.session.rollback)
        Base.metadata.create_all(self.session.bind)

    def test_abstract(self):
        with ShouldRaise(TypeError(
                "Can't instantiate abstract class SQLAlchemyDiff with "
                "abstract methods extract_imported"
        )):
            SQLAlchemyDiff([], [])

    def test_simple(self):
        self.session.add(Simple(key='a', value=1))
        self.session.add(Simple(key='b', value=2))
        self.session.add(Simple(key='c', value=3))

        imported = [
            dict(key='b', value=2),
            dict(key='c', value=4),
            dict(key='d', value=5),
        ]

        class TestDiff(SQLAlchemyDiff):

            def __init__(self, session, imported):
                existing = session.query(Simple)
                super(TestDiff, self).__init__(
                    session, Simple, existing, imported
                )

            extract_imported = MultiKeyDictExtractor('key')

        diff = TestDiff(self.session, imported)

        diff.apply()

        expected = [
            dict(key='b', value=2),
            dict(key='c', value=4),
            dict(key='d', value=5),
        ]

        actual = [dict(key=o.key, value=o.value)
                  for o in self.session.query(Simple).order_by('key')]

        compare(expected, actual)

    def test_multi_column_primary_key(self):
        self.session.add(MultiPK(name='a', index=0, value=1))
        self.session.add(MultiPK(name='b', index=0, value=2))
        self.session.add(MultiPK(name='b', index=1, value=-2))
        self.session.add(MultiPK(name='c', index=0, value=3))

        imported = [
            dict(name='b', index=0, value=2),
            dict(name='b', index=2, value=-1),
            dict(name='c', index=0, value=4),
            dict(name='d', index=0, value=5),
        ]

        class TestDiff(SQLAlchemyDiff):

            def __init__(self, session, imported):
                existing = session.query(MultiPK)
                super(TestDiff, self).__init__(
                    session, MultiPK, existing, imported
                )

            extract_imported = MultiKeyDictExtractor('name', 'index')

        diff = TestDiff(self.session, imported)

        diff.apply()

        expected = [
            dict(name='b', index=0, value=2),
            dict(name='b', index=2, value=-1),
            dict(name='c', index=0, value=4),
            dict(name='d', index=0, value=5),
        ]

        actual = [dict(name=o.name, index=o.index, value=o.value)
                  for o in self.session.query(MultiPK).order_by('name')]

        compare(expected, actual)



from collections import namedtuple
from mock import Mock, call
from unittest import TestCase
from mortar_import.extractors import DictExtractor, NamedTupleExtractor
from testfixtures import compare, ShouldRaise
from mortar_import.diff import Diff


class TestPlain(TestCase):

    def test_abstract(self):
        with ShouldRaise(TypeError(
                "Can't instantiate abstract class Diff with abstract methods "
                "add, delete, extract_existing, extract_imported, update"
        )):
            Diff([], [])

    def make_differ(self):
        mock = Mock()

        class DiffTuple(Diff):
            def extract_existing(self, obj):
                return obj[0], (obj[0], obj[-1])

            extract_imported = extract_existing

            add = mock.add
            update = mock.update
            delete = mock.delete

        return DiffTuple, mock

    def test_tuple(self):

        DiffTuple, mock = self.make_differ()

        diff = DiffTuple(
            [('a', 1, 2), ('b', 3, 4), ('c', 5, 6)],
            [('b', 3, 4), ('c', 5, 7), ('d', 7, 8)]
        )

        compare([], mock.mock_calls)

        diff.apply()

        compare([
            call.add('d', ('d', 7, 8), ('d', 8)),
            call.update('c', ('c', 5, 6), ('c', 6), ('c', 5, 7), ('c', 7)),
            call.delete('a', ('a', 1, 2), ('a', 2)),
        ], mock.mock_calls)

    def test_compute(self):

        DiffTuple, mock = self.make_differ()

        diff = DiffTuple(
            [('a', 1, 2), ('b', 3, 4), ('c', 5, 6)],
            [('b', 3, 4), ('c', 5, 7), ('d', 7, 8)]
        )

        diff.compute()

        compare([('d', ('d', 7, 8), ('d', 8))],
                diff.to_add)
        compare([('c', ('c', 5, 6), ('c', 6), ('c', 5, 7), ('c', 7))],
                diff.to_update)
        compare([('a', ('a', 1, 2), ('a', 2))],
                diff.to_delete)

        compare([], mock.mock_calls)

    def test_duplicate_existing_key(self):

        DiffTuple, mock = self.make_differ()

        diff = DiffTuple([('a', 1, 2), ('a', 3, 4)], [])

        with ShouldRaise(
                KeyError("'a' occurs more than once in existing, "
                         "first was ('a', 2) from ('a', 1, 2), "
                         "next was ('a', 4) from ('a', 3, 4)")):
            diff.apply()

    def test_duplicate_imported_key(self):

        DiffTuple, mock = self.make_differ()

        diff = DiffTuple([], [('a', 1, 2), ('a', 3, 4)])

        with ShouldRaise(
                KeyError("'a' occurs more than once in imported, "
                         "first was ('a', 2) from ('a', 1, 2), "
                         "next was ('a', 4) from ('a', 3, 4)")):
            diff.apply()

    def test_dict(self):
        mock = Mock()

        class DiffTuple(Diff):

            extract_imported = extract_existing = DictExtractor('k')

            add = mock.add
            update = mock.update
            delete = mock.delete

        a = dict(k='a', v=1)
        b = dict(k='b', v=2)
        c = dict(k='c', v=3)
        c_ = dict(k='c', v=4)
        d = dict(k='d', v=5)

        diff = DiffTuple([a, b, c], [b, c_, d])
        diff.apply()

        compare([
            call.add('d', d, d),
            call.update('c', c, c, c_, c_),
            call.delete('a', a, a),
        ], mock.mock_calls)

    def test_dict_multi_key(self):
        mock = Mock()

        class DiffTuple(Diff):

            extract_imported = extract_existing = DictExtractor('k', 'k2')

            add = mock.add
            update = mock.update
            delete = mock.delete

        a = dict(k='a', k2=0, v=1)
        b = dict(k='b', k2=0, v=2)
        c = dict(k='c', k2=0, v=3)
        c_ = dict(k='c', k2=0, v=4)
        d = dict(k='d', k2=0, v=5)

        diff = DiffTuple([a, b, c], [b, c_, d])
        diff.apply()

        compare([
            call.add(('d', 0), d, d),
            call.update(('c', 0), c, c, c_, c_),
            call.delete(('a', 0), a, a),
        ], mock.mock_calls)

    def test_named_tuple(self):
        mock = Mock()

        X = namedtuple('X', 'foo bar')
        Y = namedtuple('Y', 'foo bar')

        class DiffTuple(Diff):

            extract_imported = extract_existing = NamedTupleExtractor('foo')

            add = mock.add
            update = mock.update
            delete = mock.delete

        aX = X('a', 1)
        bX = X('b', 2)
        cX = X('c', 3)
        bY = Y('b', 2)
        cY = Y('c', 4)
        dY = Y('d', 5)

        daX = dict(foo='a', bar=1)
        dcX = dict(foo='c', bar=3)
        dcY = dict(foo='c', bar=4)
        ddY = dict(foo='d', bar=5)

        diff = DiffTuple([aX, bX, cX], [bY, cY, dY])
        diff.apply()

        compare([
            call.add(('d', ), dY, ddY),
            call.update(('c',), cX, dcX, cY, dcY),
            call.delete(('a',), aX, daX),
        ], mock.mock_calls)
